//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
//
// Maze generation: depth-first search algorithm
//
// -------------------------------------------------------------------
//
// $Id: buildD.cxx,v 1.1 2001/10/05 11:57:40 loreti Exp $

#include <iostream>
#include "maze.hh"
#include "util.hh"

using std::make_pair;
using std::vector;

// Algorithm: a room may be visited or unvisited; initially all the
// rooms are unvisited.  A starting room is chosen randomly; and, from
// there, a connection is opened to a near unvisited room (randomly
// chosen if more than one such room is found), remembering the
// sequence of visited rooms.  That process is iterated until the last
// visited room has no unvisited neighbors; then we step to the
// previous room in the sequence, and we try to iterate again from
// there.  The building process ends when the sequence of visited
// rooms becomes empty.

enum status { UNVISITED = room::UNDEFINED, VISITED };

void maze::buildD() {
  std::cout << "Depth-first search algorithm requested to generate a "
            << _width << "-by-" << _height << " maze" << std::endl;

  // The container for the sequence of visited rooms

  vector<room *> * vVR = new vector<room *>;

  // The first room, randomly chosen

  room *pR = _room + randInt(0, _roomDim);

  for ( ; ; ) {
    // Iterates.  If no unvisited room is found, removes the last one
    // from vVR and retries --- until vVR becomes empty.

    if ((pR = DFSlookup(pR, vVR)) == 0) {
      vVR->pop_back();
      if (vVR->empty()) break;
      pR = vVR->back();
    }
  }

  delete vVR;
}

room * maze::DFSlookup(
  room           *pR,
  vector<room *> *vVR
) {
  // - If the room pointed to by pR was not visited before, it is both
  //   marked as visited and added to the end of the container of
  //   visited room, vVR;
  // - the maze is searched for an unvisited room near to that one;
  // - if one is found, a road to it is opened and a pointer to it
  //   returned;
  // - if none is found, NULL is returned.

  if (pR->getFlag() == UNVISITED) {
    pR->setFlag(VISITED);
    vVR->push_back(pR);
  }

  int x = pR->getX();
  int y = pR->getY();

  vector<neighbor> next;

  // Explores the four neighbor rooms, storing the candidates for the
  // next move inside the container "next"

  wall *pW = getUpWall(x, y);
  if (pW->getType() == wall::CLOSE) {
    pR = getRoom(x, y-1);
    if (pR->getFlag() == UNVISITED) {
      next.push_back( make_pair(pR, pW) );
    }
  }

  pW = getDownWall(x, y);
  if (pW->getType() == wall::CLOSE) {
    pR = getRoom(x, y+1);
    if (pR->getFlag() == UNVISITED) {
      next.push_back( make_pair(pR, pW) );
    }
  }

  pW = getLeftWall(x, y);
  if (pW->getType() == wall::CLOSE) {
    pR = getRoom(x-1, y);
    if (pR->getFlag() == UNVISITED) {
      next.push_back( make_pair(pR, pW) );
    }
  }

  pW = getRightWall(x, y);
  if (pW->getType() == wall::CLOSE) {
    pR = getRoom(x+1, y);
    if (pR->getFlag() == UNVISITED) {
      next.push_back( make_pair(pR, pW) );
    }
  }

  if (next.empty()) {
    return 0;
  } else {
    // Choose randomly from the set of unvisited neighbor rooms

    neighbor n = next[ randInt(0, next.size()) ];
    n.second->setType( wall::OPEN );
    return n.first;
  }
}
