//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
//
// Maze generation: Prim algorithm
//
// -------------------------------------------------------------------
//
// $Id: buildP.cxx,v 1.2 2005/01/21 07:50:24 loreti Exp $

#include <algorithm>
#include <iostream>
#include "maze.hh"
#include "util.hh"

using std::make_pair;
using std::vector;

// Algorithm: a room may be "inside" the set of already visited rooms,
// "outside" that set, or on the "border".  At the beginning, all the
// rooms are labeled "outside"; one is randomly chosen and labeled
// "inside", and its neighbors labeled "border".  We choose a random
// room from the set of "border" ones, and we open a path to that room
// from a neighbor "inside" room; we label "inside" the current room;
// and all the "outside" neighbors of the current room become "border"
// rooms.  The process is iterated until there are no more "border"
// rooms.

enum status { OUTSIDE = room::UNDEFINED, BORDER, INSIDE };

void maze::buildP() {
  std::cout << "Prim algorithm requested to generate a "
            << _width << "-by-" << _height << " maze" << std::endl;

  // A container holding all the border rooms

  vector<room *> * vBR = new vector<room *>;

  // The first room

  room *pR = _room + randInt(0, _roomDim);
  primSetNeighbor(pR, vBR);

  while (!vBR->empty()) {
    // Picks up a random border room, that is moved to the end of the
    // container

    vector<room *>::iterator it   = vBR->begin() + randInt(0, vBR->size());
    vector<room *>::iterator last = vBR->end() - 1;
    std::swap(*it, *last);

    // Labels the last room "inside"; removes the room from the
    // container; iterates

    pR = vBR->back();
    vBR->pop_back();
    primSetNeighbor(pR, vBR);
  }

  delete vBR;
}

void maze::primSetNeighbor(
  room           * pR,
  vector<room *> * vBR
) {
  int x = pR->getX();
  int y = pR->getY();

  // 1) if room (x,y) is labeled BORDER, opens a way
  // 2) labels room (x,y) INSIDE
  // 3) labels all the OUTSIDE neighbors of (x,y) to BORDER, and
  //    inserts them in the vBR container

  vector<neighbor> in;
  vector<neighbor> out;

  wall *pW = getUpWall(x, y);
  if (pW->getType() == wall::CLOSE)
    primGetNeighbor(getRoom(x, y-1), pW, in, out);

  pW = getDownWall(x, y);
  if (pW->getType() == wall::CLOSE)
    primGetNeighbor(getRoom(x, y+1), pW, in, out);

  pW = getLeftWall(x, y);
  if (pW->getType() == wall::CLOSE)
    primGetNeighbor(getRoom(x-1, y), pW, in, out);

  pW = getRightWall(x, y);
  if (pW->getType() == wall::CLOSE)
    primGetNeighbor(getRoom(x+1, y), pW, in, out);

  if (pR->getFlag() == BORDER) {
    in[ randInt(0, in.size()) ].second->setType( wall::OPEN );
  }
  pR->setFlag(INSIDE);

  vector<neighbor>::iterator it, finis = out.end();
  for (it = out.begin();   it != finis;   ++it) {
    pR = it->first;
    pR->setFlag(BORDER);
    vBR->push_back(pR);
  }
}

void maze::primGetNeighbor(
  room * pR,
  wall * pW,
  vector<neighbor> & in,
  vector<neighbor> & out
) {
  // Adds a neighbor to the container of the inside or of the outside
  // ones

  switch (pR->getFlag()) {
    case INSIDE:
      in.push_back( make_pair(pR, pW) );
      break;
    case OUTSIDE:
      out.push_back( make_pair(pR, pW) );
      break;
  }
}

