//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
//
// Maze generation: class methods
//
// -------------------------------------------------------------------
//
// $Id: maze.cxx,v 1.2 2001/12/29 21:06:21 loreti Exp $

#include <iostream>
#include <stdexcept>
#include "maze.hh"
#include "util.hh"

using std::cerr;
using std::cout;
using std::runtime_error;

// Methods for the class maze

maze::maze(
  unsigned int w,
  unsigned int h
) : _width(w), _height(h), _rowFac(2*w+1),
    _roomDim(h*w), _wallDim(_rowFac*h+w)
{
  int x, y;

  _room = new room[ _roomDim ];
  _wall = new wall[ _wallDim ];

  // First of all, create all the rooms

  room *pR = _room;

  for (y = 0;   y < _height;   ++y) {
    wall *u = _wall + _rowFac * y;
    wall *d = u + _rowFac;
    wall *l = u + _width;
    wall *r = l + 1;

    for (x = 0;   x < _width;   ++x) {
      *pR++ = room(x, y, u++, d++, l++, r++);
    }
  }

  // Now, create all the walls: first, the upper line of border walls

  wall *pW = _wall;
  pR = _room;
  for (x = 0;   x < _width;   ++x) {
    *pW++ = wall(wall::BORDER, 0, pR++);
  }

  // Then all the lines of rooms with their down walls

  room *save = 0;
  pR = _room;

  for (y = 0;   y < _height;   ) {
    save = pR;
    room *prev = 0;
    wall::wallType wt = wall::BORDER;

    for (x = 0;   x < _width;  ++x) {
      *pW++ = wall(wt, prev, pR);
      prev  = pR++;
      wt    = wall::CLOSE;
    }
    *pW++ = wall(wall::BORDER, prev, 0);

    if (++y == _height) break;

    for (x = 0;   x < _width;   ++x) {
      *pW++ = wall(wt, save++, pR++);
    }
    pR = save;
  }

  // And, last, the lower border walls

  for (x = 0;   x < _width;   ++x) {
    *pW++ = wall(wall::BORDER, save++, 0);
  }
}

void maze::build(
  const buildMethod howTo,
  const entrySide   entry
) {
  int x, y;
  wall *pW;

  // Builds the internal connections for a "perfect" maze: i.e. a maze
  // where there is one and only one path connecting any pair of
  // rooms.

  switch (howTo) {
    case DFS:
      buildD();
      break;
    case PRIM:
      buildP();
      break;
    case KRUSKAL:
      buildK();
      break;
    default:
      throw runtime_error("bad algorithm");
  }

  // Opens two doors in opposite external walls, the two horizontal
  // ones or the two vertical ones; since the maze is "perfect", a
  // path (and only one) must exist that connects these doors.

  switch (entry) {
    case HORIZONTAL:
      x  = randInt(0, _width);
      y  = 0;
      pW = getUpWall(x, y);
      pW->setType(wall::OPEN);

      x  = randInt(0, _width);
      y  = _height - 1;
      pW = getDownWall(x, y);
      pW->setType(wall::OPEN);
      break;

    case VERTICAL:
      x  = 0;
      y  = randInt(0, _height);
      pW = getLeftWall(x, y);
      pW->setType(wall::OPEN);

      x  = _width - 1;
      y  = randInt(0, _height);
      pW = getRightWall(x, y);
      pW->setType(wall::OPEN);
      break;

    default:
      throw runtime_error("bad entry side");
  }
}

void maze::print(
) const {
  // An ASCII rendering of the maze

  wall *pW = _wall;
  wall::wallType wt;
  int x, y;
  char c;

  cout << "\n ";
  for (y = 0;   y <= _width;   ++y) {
    c = '*';

    for (x = 0;   x < _width;   ++x,  ++pW) {
      switch (wt = pW->getType()) {
        case wall::BORDER:
          cout << c << "***";
          break;
        case wall::OPEN:
          cout << c << "   ";
          break;
        case wall::CLOSE:
          cout << c << "---";
          break;
        default:
          cerr << "Bad wall type " << wt
               << " for horizontal wall at ("
               << x << ',' << y << ")\n";
          throw runtime_error("Bad maze");
      }
      if (y > 0   &&   y < _height) c = '+';
    }
    cout << "*\n ";

    if (y == _height) break;

    for (x = 0;   x < _width;   ++x,  ++pW) {
      switch (wt = pW->getType()) {
        case wall::BORDER:
          cout << "*   ";
          break;
        case wall::OPEN:
          cout << "    ";
          break;
        case wall::CLOSE:
          cout << "|   ";
          break;
        default:
          cerr << "Bad wall type " << wt
               << " for vertical wall at ("
               << x << ',' << y << ")\n";
          throw runtime_error("Bad maze");
      }
    }

    switch (wt = (pW++)->getType()) {
      case wall::OPEN:
        cout << " \n ";
        break;
      case wall::BORDER:
        cout << "*\n ";
        break;
      default:
        cerr << "Bad wall type " << wt
             << " for vertical wall at ("
             << _width << ',' << y << ")\n";
        throw runtime_error("Bad maze");
    }
  }
  cout << std::endl;
}
